import logging
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity

logger = logging.getLogger("MaximSDK")


@dataclass
class ToolCallConfig:
    id: str
    name: str
    description: str
    args: str
    tags: Optional[Dict[str, str]] = None


@dataclass
class ToolCallError:
    message: str
    code: Optional[str] = None
    type: Optional[str] = None


class ToolCall(BaseContainer):
    def __init__(self, config: ToolCallConfig, writer: LogWriter):
        super().__init__(Entity.TOOL_CALL, config.__dict__, writer)
        self._id = config.id
        self._name = config.name
        self.args = config.args
        self.description = config.description
        self.tags = config.tags

    def update(self, data: Dict[str, Any]):
        self._commit("update", data)        
        
    @staticmethod
    def update_(writer: LogWriter, id: str, data: Dict[str, Any]):
        BaseContainer._commit_(writer, Entity.TOOL_CALL, id, "update", data)

    @staticmethod
    def result_(writer: LogWriter, id: str, result: str):
        BaseContainer._commit_(
            writer, Entity.TOOL_CALL, id, "result", {"result": result}
        )
        BaseContainer._end_(
            writer,
            Entity.TOOL_CALL,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def attach_evaluators(self, evaluators: List[str]):
        raise NotImplementedError("attach_evaluators is not supported for ToolCall")

    def with_variables(self, for_evaluators: List[str], variables: Dict[str, str]):
        raise NotImplementedError("with_variables is not supported for ToolCall")

    def result(self, result: str):
        self._commit("result", {"result": result})
        self.end()

    def error(self, error: ToolCallError):
        self._commit("error", {"error": error})
        self.end()

    @staticmethod
    def error_(writer: LogWriter, id: str, error: ToolCallError):
        BaseContainer._commit_(writer, Entity.TOOL_CALL, id, "error", {"error": error})
        BaseContainer._end_(
            writer,
            Entity.TOOL_CALL,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        return {
            **base_data,
            "name": self._name,
            "description": self.description,
            "args": self.args,
        }



import json
from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Union


class Entity(Enum):
    SESSION = "session"
    TRACE = "trace"
    SPAN = "span"
    TOOL_CALL = "tool_call"
    GENERATION = "generation"
    FEEDBACK = "feedback"
    RETRIEVAL = "retrieval"


class DateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return super().default(o)


class CommitLog:

    def __init__(
        self,
        entity: Entity,
        entity_id: str,
        action: str,
        data: Optional[Dict[str, Any]] = None,
    ):
        self.entity = entity
        self.entity_id = entity_id
        self.action = action
        self.data = data

    def serialize(self, custom_data: Optional[Dict[str, Any]] = None) -> str:
        if custom_data is not None:
            if self.data is None:
                self.data = {}
            self.data.update(custom_data)
        return f"{self.entity.value}{{id={self.entity_id},action={self.action},data={json.dumps(self.data,cls=DateTimeEncoder)}}}"


def object_to_dict(obj: Any) -> Union[Dict, List, str, int, float, bool, None]:
    """
    Convert a complex object structure to a dictionary, handling nested custom objects.

    Args:
        obj: Any Python object to convert

    Returns:
        A dictionary representation of the object
    """
    if obj is None:
        return None

    # Handle basic types
    if isinstance(obj, (str, int, float, bool)):
        return obj

    # Handle datetime objects
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()

    # Handle dictionaries
    if isinstance(obj, dict):
        return {k: object_to_dict(v) for k, v in obj.items()}

    # Handle lists and tuples
    if isinstance(obj, (list, tuple)):
        return [object_to_dict(item) for item in obj]

    # Handle SimpleNamespace
    if isinstance(obj, SimpleNamespace):
        return object_to_dict(vars(obj))

    # Handle custom objects with __dict__
    if hasattr(obj, "__dict__"):
        # Get all attributes, including properties
        attrs = {}
        for key in dir(obj):
            # Skip private attributes and methods
            if not key.startswith("_"):
                value = getattr(obj, key)
                # Skip callable attributes (methods)
                if not callable(value):
                    attrs[key] = object_to_dict(value)
        return attrs

    # Handle other types by converting to string
    return str(obj)


@dataclass
class GenerationError:
    message: str
    code: Optional[str] = None
    type: Optional[str] = None



import asyncio
from typing import Awaitable, Dict, List, Union

from maxim.evaluators.base_evaluator import BaseEvaluator

from ..apis.maxim_apis import MaximAPI
from ..models.dataset import ManualData
from ..models.evaluator import (
    LocalEvaluationResult,
    LocalEvaluatorResultParameter,
    LocalEvaluatorReturn,
)
from ..models.test_run import YieldedOutput, YieldedOutputMeta, YieldedOutputTokenUsage


async def process_awaitable(awaitable: Awaitable):
    return await awaitable


async def run_local_evaluations(
    evaluators: List[BaseEvaluator],
    data_entry: ManualData,
    processed_data: LocalEvaluatorResultParameter,
) -> List[LocalEvaluationResult]:
    coroutines = [
        asyncio.to_thread(evaluator.guarded_evaluate, processed_data, data_entry)
        for evaluator in evaluators
    ]
    evaluator_results = await asyncio.gather(*coroutines)
    results: List[LocalEvaluationResult] = []
    for i, evaluator in enumerate(evaluators):
        if isinstance(evaluator, BaseEvaluator):
            try:
                combined_results = evaluator_results[i]
                for name, result in combined_results.items():
                    results.append(LocalEvaluationResult(
                        name=name,
                        pass_fail_criteria=evaluator.pass_fail_criteria[name],
                        result=result
                    ))
            except Exception as err:
                results.extend(
                    [
                        LocalEvaluationResult(
                            name=name,
                            pass_fail_criteria=evaluator.pass_fail_criteria[name],
                            result=LocalEvaluatorReturn(
                                score="Err",
                                reasoning=f"Error while running combined evaluator with names {evaluator.names}: {str(err)}",
                            ),
                        )
                        for name in evaluator.names
                    ]
                )
    return results


def workflow_id_output_function_closure(workflow_id: str, api_service: MaximAPI):
    async def output_function(
        data: Dict[str, Union[str, List[str], None]],
    ) -> YieldedOutput:
        result = api_service.execute_workflow_for_data(
            workflow_id=workflow_id, data_entry=data
        )

        output: str
        if result.output is None:
            output = ""
        else:
            output = result.output

        return YieldedOutput(
            data=output,
            retrieved_context_to_evaluate=result.context_to_evaluate,
            meta=YieldedOutputMeta(
                usage=YieldedOutputTokenUsage(
                    latency=result.latency,
                    total_tokens=0,
                    prompt_tokens=0,
                    completion_tokens=0,
                )
            ),
        )

    return output_function


def prompt_version_id_output_function_closure(
    prompt_version_id: str, input_data: str, api_service: MaximAPI
):
    async def output_function(
        data: Dict[str, Union[str, List[str], None]],
    ) -> YieldedOutput:
        result = api_service.execute_prompt_for_data(
            prompt_version_id=prompt_version_id, input=input_data, data_entry=data
        )

        output: str
        if result.output is None:
            output = ""
        else:
            output = result.output

        return YieldedOutput(
            data=output,
            retrieved_context_to_evaluate=result.context_to_evaluate,
            meta=YieldedOutputMeta(
                cost=result.cost if result.cost is not None else None,
                usage=result.usage if result.usage is not None else None,
            ),
        )

    return output_function



import logging
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Dict, Iterator, List, Optional, Union
from uuid import uuid4

from ..parsers.generation_parser import parse_model_parameters, parse_result
from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity, GenerationError, object_to_dict


@dataclass
class GenerationRequestMessage:
    role: str
    content: str

    def to_dict(self) -> Dict[str, str]:
        return {"role": self.role, "content": self.content}

    @classmethod
    def from_gemini_content(cls, content: Any) -> "GenerationRequestMessage":
        if "role" not in content or "parts" not in content:
            raise ValueError("[MaximSDK] Invalid Gemini content")
        if not isinstance(content["parts"], list):
            raise ValueError("[MaximSDK] Invalid parts in Gemini content.")
        parts_content = ""
        for part in content["parts"]:
            parts_content += part["text"]
        return GenerationRequestMessage(role=content["role"], content=parts_content)


@dataclass
class GenerationConfig:
    id: str
    provider: str
    model: str
    messages: Optional[List[GenerationRequestMessage]] = field(default_factory=list)
    model_parameters: Dict[str, Any] = field(default_factory=dict)
    span_id: Optional[str] = None
    name: Optional[str] = None
    maxim_prompt_id: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


valid_providers = [
    "openai",
    "azure",
    "anthropic",
    "huggingface",
    "together",
    "google",
    "groq",
    "bedrock",
    "cohere",
    "unknown",
]


@dataclass
class GenerationToolCallFunction:
    name: str
    arguments: Optional[str]


@dataclass
class GenerationToolCall:
    id: str
    type: str
    function: GenerationToolCallFunction


@dataclass
class GenerationResultMessage:
    role: str
    content: Optional[str]
    tool_calls: Optional[List[GenerationToolCall]]


@dataclass
class GenerationResultChoice:
    index: int
    message: GenerationResultMessage
    logprobs: Optional[Any] = None
    finish_reason: Optional[str] = None


@dataclass
class GenerationUsage:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


@dataclass
class GenerationResult:
    id: str
    object: str
    created: int
    model: str
    choices: List[GenerationResultChoice]
    usage: GenerationUsage


class Generation(BaseContainer):
    def __init__(self, config: GenerationConfig, writer: LogWriter):
        super().__init__(Entity.GENERATION, config.__dict__, writer)
        self.model = config.model
        self.maxim_prompt_id = config.maxim_prompt_id
        self.messages = []
        self.provider = config.provider
        self.provider = config.provider.lower()
        if self.provider not in valid_providers:
            self.provider = "unknown"
        if config.messages is not None:
            self.messages.extend([m.to_dict() for m in config.messages])
        self.model_parameters = parse_model_parameters(config.model_parameters)

    def _set_provider(self, writer: LogWriter, id: str, provider: str):
        if provider not in valid_providers:
            raise ValueError(
                f"Invalid provider: {self.provider}. Must be one of {', '.join(valid_providers)}."
            )
        BaseContainer._commit_(
            writer, Entity.GENERATION, id, "update", {"provider": provider}
        )

    def set_provider(self, provider: str):
        if provider not in valid_providers:
            raise ValueError(
                f"Invalid provider: {self.provider}. Must be one of {', '.join(valid_providers)}."
            )
        self.provider = provider

    @staticmethod
    def set_model_(writer: LogWriter, id: str, model: str):
        BaseContainer._commit_(
            writer, Entity.GENERATION, id, "update", {"model": model}
        )

    def set_model(self, model: str):
        self.model = model
        self._commit("update", {"model": model})

    @staticmethod
    def add_message_(writer: LogWriter, id: str, message: GenerationRequestMessage):
        BaseContainer._commit_(
            writer, Entity.GENERATION, id, "update", {"messages": [message.to_dict()]}
        )

    def add_message(self, message: GenerationRequestMessage):
        self.messages.append(message)
        self._commit("update", {"messages": [message.to_dict()]})

    @staticmethod
    def set_model_parameters_(
        writer: LogWriter, id: str, model_parameters: Dict[str, Any]
    ):
        BaseContainer._commit_(
            writer,
            Entity.GENERATION,
            id,
            "update",
            {"model_parameters": model_parameters},
        )

    def set_model_parameters(self, model_parameters: Dict[str, Any]):
        self.model_parameters = model_parameters
        self._commit("update", {"model_parameters": model_parameters})

    @staticmethod
    def result_(
        writer: LogWriter, id: str, result: Union[GenerationResult, Dict[str, Any]]
    ):
        try:
            # Checking the type
            result = Generation.convert_result(result)
            # Validating the result
            parse_result(result)
            BaseContainer._commit_(
                writer, Entity.GENERATION, id, "result", {"result": result}
            )
            BaseContainer._end_(
                writer,
                Entity.GENERATION,
                id,
                {
                    "endTimestamp": datetime.now(timezone.utc),
                },
            )
        except Exception as e:
            logging.error(
                f"[MaximSDK] Invalid result. You can pass OpenAI/Azure ChatCompletion or Langchain LLMResult,AIMessage,ToolMessage or LiteLLM ModelResponse: {str(e)}"
            )

    @staticmethod
    def end_(writer: LogWriter, id: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}
        BaseContainer._end_(
            writer,
            Entity.GENERATION,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
                **data,
            },
        )

    @staticmethod
    def add_tag_(writer: LogWriter, id: str, key: str, value: str):
        BaseContainer._add_tag_(writer, Entity.GENERATION, id, key, value)

    @staticmethod
    def convert_chat_completion(chat_completion: Dict[str, Any]):
        return {
            "id": chat_completion.get("id", str(uuid4())),
            "created": chat_completion.get("created", datetime.now(timezone.utc)),
            "choices": [
                {
                    "index": choice.get("index", 0),
                    "message": {
                        "role": choice.get("message").get("role", "assistant"),
                        "content": choice.get("message").get("content", ""),
                        "tool_calls": choice.get("message").get("tool_calls", None),
                        "function_calls": choice.get("message").get(
                            "function_calls", None
                        ),
                    },
                    "finish_reason": choice.get("finish_reason", None),
                    "logprobs": choice.get("logprobs", None),
                }
                for choice in chat_completion.get("choices", [])
            ],
            "usage": chat_completion.get("usage", {}),
        }

    @staticmethod
    def convert_result(
        result: Union[Any, GenerationResult, Dict[str, Any]],
    ) -> Union[Any, GenerationResult, Dict[str, Any]]:
        try:
            parse_result(result)
            return result
        except Exception:
            if isinstance(result, object):
                # trying for langchain first
                try:
                    from langchain_core.messages import AIMessage  # type: ignore
                    from langchain_core.outputs import LLMResult  # type: ignore

                    from ..langchain.utils import (
                        parse_base_message_to_maxim_generation,
                        parse_langchain_llm_result,
                    )

                    if isinstance(result, AIMessage):
                        return parse_base_message_to_maxim_generation(result)
                    elif isinstance(result, LLMResult):
                        return parse_langchain_llm_result(result)
                except ImportError:
                    pass
                # trying for litellm
                try:
                    from litellm.types.utils import ModelResponse  # type: ignore

                    from ..litellm.parser import parse_litellm_model_response

                    if isinstance(result, ModelResponse):
                        return parse_litellm_model_response(result)
                except ImportError:
                    pass
                # trying gemini response
                try:
                    from google.genai.types import GenerateContentResponse

                    from ..gemini.utils import GeminiUtils

                    if isinstance(result, GenerateContentResponse):
                        return GeminiUtils.parse_gemini_generation_content(result)
                    elif isinstance(result, Iterator):
                        return GeminiUtils.parse_gemini_generation_content_iterator(
                            result
                        )
                except ImportError:
                    pass
            result_dict = object_to_dict(result)
            if isinstance(result_dict, Dict):
                # Checking if its Azure or OpenAI result
                if (
                    "object" in result_dict
                    and result_dict["object"] == "chat.completion"
                ):
                    return Generation.convert_chat_completion(result_dict)
                elif (
                    "object" in result_dict
                    and result_dict["object"] == "text.completion"
                ):
                    raise ValueError("Text completion is not yet supported.")

            return result

    def result(self, result: Any):
        try:
            # Checking the type
            result = Generation.convert_result(result)
            # Validating the result
            parse_result(result)
            # Logging the result
            self._commit("result", {"result": result})
            self.end()
        except ValueError as e:
            logging.error(
                f"[MaximSDK] Invalid result. You can pass OpenAI/Azure ChatCompletion or Langchain LLMResult,AIMessage,ToolMessage or LiteLLM ModelResponse: {str(e)}"
            )

    def error(self, error: GenerationError):
        if not error.code:
            error.code = ""
        if not error.type:
            error.type = ""
        self._commit(
            "result",
            {
                "result": {
                    "error": {
                        "message": error.message,
                        "code": error.code,
                        "type": error.type,
                    },
                    "id": str(uuid4()),
                }
            },
        )
        self.end()

    @staticmethod
    def error_(writer: LogWriter, id: str, error: GenerationError):
        if not error.code:
            error.code = ""
        if not error.type:
            error.type = ""
        BaseContainer._commit_(
            writer,
            Entity.GENERATION,
            id,
            "result",
            {
                "result": {
                    "error": {
                        "message": error.message,
                        "code": error.code,
                        "type": error.type,
                    },
                    "id": str(uuid4()),
                }
            },
        )
        BaseContainer._end_(
            writer,
            Entity.GENERATION,
            id,
            {
                "endTimestamp": datetime.now(timezone.utc),
            },
        )

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        return {
            **base_data,
            "model": self.model,
            "provider": self.provider,
            "maximPromptId": self.maxim_prompt_id,
            "messages": self.messages,
            "modelParameters": self.model_parameters,
        }



import logging

logger = logging.getLogger("MaximSDK")

try:
    import litellm

    LiteLLMAvailable = True
except ImportError:
    LiteLLMAvailable = False

if not LiteLLMAvailable:
    logger.error("LiteLLM is not available. You can't use MaximLiteLLMTracer.")

from .tracer import MaximLiteLLMTracer

__all__ = ["MaximLiteLLMTracer"]



import logging
from typing import Any, Dict, Optional
from uuid import uuid4

from litellm.integrations.custom_logger import CustomLogger

from ..logger import GenerationConfig, GenerationError, Logger
from ..models import Container, SpanContainer, TraceContainer


class MaximLiteLLMTracer(CustomLogger):
    """
    Custom logger for Litellm.
    """

    def __init__(self, logger: Logger):
        super().__init__()
        logging.warning("[MaximSDK] Litellm support is in beta")
        self.logger = logger
        self.containers: Dict[str, Container] = {}

    def __get_container_from_metadata(
        self, metadata: Optional[Dict[str, Any]]
    ) -> Container:
        if metadata is not None and metadata["trace_id"] is not None:
            trace_id = metadata["trace_id"] if "trace_id" in metadata else None
            span_name = metadata["span_name"] if "span_name" in metadata else None
            tags = metadata["span_tags"] if "span_tags" in metadata else None
            if trace_id is not None:
                # Here we will create a new span and send back that as container
                container = SpanContainer(
                    span_id=str(str(uuid4())),
                    logger=self.logger,
                    span_name=span_name,
                    parent=trace_id,
                )
                container.create()
                if tags is not None:
                    container.add_tags(tags)
                return container
            # We will be creating trace from scratch
            tags = metadata["trace_tags"] if "trace_tags" in metadata else None
            trace_name = metadata["trace_name"] if "trace_name" in metadata else None
            session_id = metadata["session_id"] if "session_id" in metadata else None
            container = TraceContainer(
                trace_id=str(str(uuid4())),
                logger=self.logger,
                trace_name=trace_name,
                parent=session_id,
            )
            container.create()
            if tags is not None:
                container.add_tags(tags)
            return container

        return TraceContainer(
            trace_id=str(uuid4()), logger=self.logger, trace_name="LiteLLM"
        )

    def log_pre_api_call(self, model, messages, kwargs):
        """
        Runs when a LLM call starts.
        """
        try:
            metadata: Optional[dict[str, Any]] = None
            generation_name = None
            tags = {}
            litellm_metadata = kwargs["litellm_params"]["metadata"] or {}
            if litellm_metadata:
                metadata = (
                    litellm_metadata["maxim"] if litellm_metadata["maxim"] else None
                )
                if metadata is not None:
                    generation_name = (
                        metadata["generation_name"]
                        if "generation_name" in metadata
                        else None
                    )
                    tags = (
                        metadata["generation_tags"]
                        if "generation_tags" in metadata
                        else None
                    )

            # checking if trace_id present in metadata
            container = self.__get_container_from_metadata(metadata)
            if not container.is_created():
                container.create()
            call_id = kwargs["litellm_call_id"]
            self.containers[call_id] = container
            # starting trace
            provider = kwargs["litellm_params"]["custom_llm_provider"]
            params: Dict[str, Any] = (
                kwargs["optional_params"] if "optional_params" in kwargs else {}
            )
            _ = container.add_generation(
                GenerationConfig(
                    id=call_id,
                    messages=messages,
                    model=model,
                    provider=provider,
                    name=generation_name,
                    tags=tags,
                    model_parameters=params,
                )
            )
        except Exception:
            logging.error(
                "[MaximSDK] Error while handling pre_api_call for litellm: {e}"
            )

    def log_success_event(self, kwargs, response_obj, start_time, end_time):
        """
        Runs when a LLM call succeeds.
        """
        try:
            call_id = kwargs["litellm_call_id"]
            container = self.containers[call_id] if call_id in self.containers else None
            if container is None:
                logging.warning(
                    "[MaximSDK] Couldn't find container for logging Litellm post call."
                )
                return
            self.logger.generation_result(call_id, result=response_obj)
            container.end()
        except Exception:
            logging.error(
                "[MaximSDK] Error while handling log_success_event for litellm: ${e}"
            )

    def log_failure_event(self, kwargs, response_obj, start_time, end_time):
        """
        Runs when a LLM call fails.
        """
        try:
            call_id = kwargs["litellm_call_id"]
            container = self.containers[call_id] if call_id in self.containers else None
            if container is None:
                # This means that this was an litellm level error
                container = self.__get_container_from_metadata(None)
                if not container.is_created():
                    container.create()
                model = kwargs["model"] if "model" in kwargs else None
                messages = kwargs["messages"] if "messages" in kwargs else None
                provider = (
                    kwargs["custom_llm_provider"]
                    if "custom_llm_providervider" in kwargs
                    else None
                )
                container.add_generation(
                    GenerationConfig(
                        id=call_id,
                        messages=messages,
                        model=model or "Unknown",
                        provider=provider or "Unknown",
                    )
                )
            exception = kwargs["exception"] or None
            if exception is not None:
                self.logger.generation_error(
                    generation_id=call_id,
                    error=GenerationError(
                        message=exception.message,
                        code=str(exception.status_code),
                    ),
                )
            container.end()
        except Exception as e:
            logging.error(
                f"[MaximSDK] Error while handling log_failure_event for litellm {e}"
            )

    async def async_log_pre_api_call(self, model, messages, kwargs):
        """
        Runs when a LLM call starts.
        """
        try:
            metadata: Optional[dict[str, Any]] = None
            generation_name = None
            tags = {}
            litellm_metadata = kwargs["litellm_params"]["metadata"] or {}
            if litellm_metadata:
                metadata = (
                    litellm_metadata["maxim"] if litellm_metadata["maxim"] else None
                )
                if metadata is not None:
                    generation_name = metadata["generation_name"]
                    tags = metadata["generation_tags"]

            # checking if trace_id present in metadata
            container = self.__get_container_from_metadata(metadata)
            if not container.is_created():
                container.create()
            call_id = kwargs["litellm_call_id"]
            self.containers[call_id] = container
            # starting trace
            provider = kwargs["litellm_params"]["custom_llm_provider"]
            model_parameters = {}
            _ = container.add_generation(
                GenerationConfig(
                    id=call_id,
                    messages=messages,
                    model=model,
                    provider=provider,
                    name=generation_name,
                    tags=tags,
                )
            )
        except Exception as e:
            logging.error(
                f"[MaximSDK] Error while handling async_log_pre_api_call for litellm: {e}"
            )

    async def async_log_success_event(self, kwargs, response_obj, start_time, end_time):
        """
        Runs when a LLM call succeeds.
        """
        try:
            call_id = kwargs["litellm_call_id"]
            container = self.containers[call_id] if call_id in self.containers else None
            if container is None:
                logging.warning(
                    "[MaximSDK] Couldn't find container for logging Litellm post call."
                )
                return
            self.logger.generation_result(call_id, result=response_obj)
            container.end()
        except Exception as e:
            logging.error(
                f"[MaximSDK] Error while handling async_log_success_event for litellm: {e}"
            )

    async def async_log_failure_event(self, kwargs, response_obj, start_time, end_time):
        """
        Runs when a LLM call fails.
        """
        try:
            call_id = kwargs["litellm_call_id"]
            container = self.containers[call_id] if call_id in self.containers else None
            if container is None:
                # This means that this was an litellm level error
                container = self.__get_container_from_metadata(None)
                if not container.is_created():
                    container.create()
                model = kwargs["model"] if "model" in kwargs else None
                messages = kwargs["messages"] if "messages" in kwargs else None
                provider = (
                    kwargs["custom_llm_provider"]
                    if "custom_llm_providervider" in kwargs
                    else None
                )
                container.add_generation(
                    GenerationConfig(
                        id=call_id,
                        messages=messages,
                        model=model or "Unknown",
                        provider=provider or "Unknown",
                    )
                )
            exception = kwargs["exception"] or None
            if exception is not None:
                self.logger.generation_error(
                    generation_id=call_id,
                    error=GenerationError(
                        message=exception.message,
                        code=str(exception.status_code),
                    ),
                )
            container.end()
        except Exception as e:
            logging.error(
                f"[MaximSDK] Error while handling async_log_failure_event for litellm: {e}"
            )



import json
import logging
import threading
import time
from dataclasses import dataclass
from typing import Dict, List, Optional

from .apis import MaximAPI
from .cache import MaximCache, MaximInMemoryCache
from .filter_objects import (
    IncomingQuery,
    QueryObject,
    find_all_matches,
    find_best_match,
    parse_incoming_query,
)
from .logger import Logger, LoggerConfig
from .models import (
    Folder,
    FolderEncoder,
    Prompt,
    PromptChain,
    PromptChainVersion,
    PromptChainVersionsAndRules,
    PromptVersion,
    QueryRule,
    RuleGroupType,
    RuleType,
    VersionAndRulesWithPromptChainIdEncoder,
    VersionAndRulesWithPromptId,
    VersionAndRulesWithPromptIdEncoder,
    VersionsAndRules,
)
from .test_runs import TestRunBuilder


@dataclass
class Config:
    """
    A class representing the configuration for the Maxim SDK.

    Attributes:
        api_key (str): The API key for the Maxim instance.
        base_url (Optional[str], optional): The base URL for the Maxim instance. Defaults to "https://app.getmaxim.ai".
        cache (Optional[MaximCache], optional): The cache to use for the Maxim instance. Defaults to None.
        debug (Optional[bool], optional): Whether to enable debug logging. Defaults to False.
        raise_exceptions (Optional[bool], optional): Whether to raise exceptions during logging operations. Defaults to False.
        prompt_management (Optional[bool], optional): Whether to enable prompt management. Defaults to False.
    """

    api_key: str
    base_url: Optional[str] = None
    cache: Optional[MaximCache] = None
    debug: Optional[bool] = False
    raise_exceptions: Optional[bool] = False
    prompt_management: Optional[bool] = False


EntityType = {"PROMPT": "PROMPT", "FOLDER": "FOLDER", "PROMPT_CHAIN": "PROMPT_CHAIN"}


class Maxim:
    """
    A class representing the Maxim SDK.

    This class provides methods for interacting with the Maxim API.
    """

    def __init__(self, config: Config):
        """
        Initializes a new instance of the Maxim class.

        Args:
            config (Config): The configuration for the Maxim instance.
        """
        self.base_url = config.base_url or "https://app.getmaxim.ai"
        self.api_key = config.api_key
        self.is_running = True
        self.raise_exceptions = config.raise_exceptions or False
        self.maxim_api = MaximAPI(self.base_url, self.api_key)
        self.__is_debug = config.debug or False
        self.__loggers: Dict[str, Logger] = {}
        self.prompt_management = config.prompt_management
        if self.prompt_management:
            self.__cache = config.cache or MaximInMemoryCache()
            self.__sync_thread = threading.Thread(target=self.__sync_timer)
            self.__sync_thread.daemon = True
            self.__sync_thread.start()

    def enable_prompt_management(self) -> "Maxim":
        self.prompt_management = True
        if not self.__sync_thread.is_alive():
            self.__sync_thread = threading.Thread(target=self.__sync_timer)
            self.__sync_thread.daemon = True
            self.__sync_thread.start()
        return self

    def enable_exceptions(self, val: bool) -> "Maxim":
        self.raise_exceptions = val
        return self

    def __sync_timer(self):
        if not self.prompt_management:
            return
        while self.is_running:
            self.__sync_entities()
            time.sleep(60)

    def __sync_entities(self):
        logging.debug(" Syncing prompts and folders")
        if not self.prompt_management:
            return
        if not self.is_running:
            return
        self.__sync_prompts()
        self.__sync_prompt_chains()
        self.__syncFolders()
        logging.debug("Syncing completed")

    def __sync_prompts(self):
        logging.debug("Syncing prompts")
        try:
            prompts = self.maxim_api.get_prompts()
            logging.debug(f"Found {len(prompts)} prompts")
            for prompt in prompts:
                try:
                    self.__cache.set(
                        self.__get_cache_key(EntityType["PROMPT"], prompt.promptId),
                        json.dumps(prompt, cls=VersionAndRulesWithPromptIdEncoder),
                    )
                except Exception as err:
                    logging.error(err)
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            logging.error(f"Error while syncing prompts: {err}")
            if self.raise_exceptions:
                raise err

    def __sync_prompt_chains(self):
        logging.debug("Syncing prompt Chains")
        try:
            prompt_chains = self.maxim_api.get_prompt_chains()
            logging.debug(f"Found {len(prompt_chains)} prompt chains")
            for prompt_chain in prompt_chains:
                try:
                    self.__cache.set(
                        self.__get_cache_key(
                            EntityType["PROMPT_CHAIN"], prompt_chain.promptChainId
                        ),
                        json.dumps(
                            prompt_chain, cls=VersionAndRulesWithPromptChainIdEncoder
                        ),
                    )
                except Exception as err:
                    logging.error(err)
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            logging.error(f"Error while syncing prompts: {err}")
            if self.raise_exceptions:
                raise err

    def __syncFolders(self):
        logging.debug("Syncing folders")
        try:
            folders = self.maxim_api.get_folders()
            logging.debug(f"Found {len(folders)} folders")
            for folder in folders:
                try:
                    self.__cache.set(
                        self.__get_cache_key(EntityType["FOLDER"], folder.id),
                        json.dumps(folder, cls=FolderEncoder),
                    )
                except Exception as err:
                    logging.error(err)
                    if self.raise_exceptions:
                        raise err
        except Exception as err:
            logging.error(f"Error while syncing folders: {err}")
            if self.raise_exceptions:
                raise err

    def __get_cache_key(self, entity: str, id: str) -> str:
        if entity == "PROMPT":
            return f"prompt:{id}"
        else:
            return f"folder:{id}"

    def __get_prompt_from_cache(self, key: str) -> Optional[VersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return VersionAndRulesWithPromptId.from_dict(json_data)

    def __get_all_prompts_from_cache(self) -> Optional[List[VersionsAndRules]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [self.__cache.get(key) for key in keys if key.startswith("prompt:")]
        prompt_list = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if "promptId" in json_data:
                    json_data.pop("promptId")
                prompt_list.append(VersionsAndRules.from_dict(json_data))
        return prompt_list

    def __get_prompt_chain_from_cache(
        self, key: str
    ) -> Optional[PromptChainVersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        parsed_data = json.loads(data)
        return PromptChainVersionsAndRules.from_dict(parsed_data)

    def __get_all_prompt_chains_from_cache(
        self,
    ) -> Optional[List[PromptChainVersionsAndRules]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [
            self.__cache.get(key) for key in keys if key.startswith("prompt_chain:")
        ]
        prompt_chain_list = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if "promptChainId" in json_data:
                    json_data.pop("promptChainId")
                prompt_chain_list.append(
                    PromptChainVersionsAndRules.from_dict(json_data)
                )
        return prompt_chain_list

    def __get_folder_from_cache(self, key: str) -> Optional[Folder]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return Folder(**json_data)

    def __get_all_folders_from_cache(self) -> Optional[List[Folder]]:
        keys = self.__cache.get_all_keys()
        if not keys:
            return None
        data = [self.__cache.get(key) for key in keys if key.startswith("folder:")]
        return [Folder(**json.loads(d)) for d in data if d is not None]

    def __format_prompt(self, prompt: PromptVersion) -> Prompt:
        json_data = {
            "promptId": prompt.promptId,
            "version": prompt.version,
            "versionId": prompt.id,
            "tags": prompt.config.tags if prompt.config else [],
            "model": prompt.config.model if prompt.config else None,
            "messages": prompt.config.messages if prompt.config else None,
            "modelParameters": prompt.config.modelParameters if prompt.config else None,
        }
        return Prompt(**json_data)

    def __format_prompt_chain(
        self, prompt_chain_version: PromptChainVersion
    ) -> PromptChain:
        json_data = {
            "promptChainId": prompt_chain_version.promptChainId,
            "versionId": prompt_chain_version.id,
            "version": prompt_chain_version.version,
            "nodes": (
                [
                    n
                    for n in prompt_chain_version.config.nodes
                    if hasattr(n, "prompt") or (isinstance(n, dict) and "prompt" in n)
                ]
                if prompt_chain_version.config
                else []
            ),
        }
        return PromptChain(**json_data)

    def __get_prompt_version_for_rule(
        self,
        prompt_version_and_rules: VersionsAndRules,
        rule: Optional[QueryRule] = None,
    ) -> Optional[Prompt]:
        if rule:
            incoming_query = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch
            )
            objects = []
            for version_id, version_rules in prompt_version_and_rules.rules.items():
                for version_rule in version_rules:
                    if not version_rule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version = next(
                        (
                            v
                            for v in prompt_version_and_rules.versions
                            if v.id == version_id
                        ),
                        None,
                    )
                    if not version:
                        continue
                    query = version_rule.rules.query
                    if version.config and version.config.tags:
                        parsed_incoming_query = parse_incoming_query(
                            incoming_query.query
                        )
                        tags = version.config.tags
                        for key, value in tags.items():
                            if value == None:
                                continue
                            if parsed_incoming_query is None:
                                continue
                            incomingQueryFields = [
                                q.field for q in parsed_incoming_query
                            ]
                            if key in incomingQueryFields:
                                query.rules.append(
                                    RuleType(field=key, operator="=", value=value)
                                )
                    objects.append(QueryObject(query=query, id=version_id))

            deployed_version_object = find_best_match(objects, incoming_query)
            if deployed_version_object:
                deployed_version = next(
                    (
                        v
                        for v in prompt_version_and_rules.versions
                        if v.id == deployed_version_object.id
                    ),
                    None,
                )
                if deployed_version:
                    return self.__format_prompt(deployed_version)

        else:
            if prompt_version_and_rules.rules:
                for version_id, version_rules in prompt_version_and_rules.rules.items():
                    logging.debug(f"type: {type(version_rules[0].rules)}")
                    does_query_exist = any(
                        ruleElm.rules.query is not None for ruleElm in version_rules
                    )
                    if does_query_exist:
                        deployed_version = next(
                            (
                                v
                                for v in prompt_version_and_rules.versions
                                if v.id == version_id
                            ),
                            None,
                        )
                        if deployed_version:
                            return self.__format_prompt(deployed_version)
            else:
                return self.__format_prompt(prompt_version_and_rules.versions[0])

        if prompt_version_and_rules.fallbackVersion:
            logging.info(
                f"************* FALLBACK TRIGGERED : VERSION : {prompt_version_and_rules.fallbackVersion} *************"
            )
            return self.__format_prompt(prompt_version_and_rules.fallbackVersion)
        return None

    def __get_prompt_chain_version_for_rule(
        self,
        prompt_chain_version_and_rules: PromptChainVersionsAndRules,
        rule: Optional[QueryRule] = None,
    ) -> Optional[PromptChain]:
        if rule:
            incoming_query = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch
            )
            objects = []
            for (
                version_id,
                version_rules,
            ) in prompt_chain_version_and_rules.rules.items():
                for version_rule in version_rules:
                    if not version_rule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version = next(
                        (
                            v
                            for v in prompt_chain_version_and_rules.versions
                            if v.id == version_id
                        ),
                        None,
                    )
                    if not version:
                        continue
                    query = RuleGroupType(**version_rule.rules.query.__dict__)
                    objects.append(QueryObject(query=query, id=version_id))

            deployed_version_object = find_best_match(objects, incoming_query)
            if deployed_version_object:
                deployed_version = next(
                    (
                        v
                        for v in prompt_chain_version_and_rules.versions
                        if v.id == deployed_version_object.id
                    ),
                    None,
                )
                if deployed_version:
                    return self.__format_prompt_chain(deployed_version)

        else:
            if prompt_chain_version_and_rules.rules:
                for (
                    version_id,
                    version_rules,
                ) in prompt_chain_version_and_rules.rules.items():
                    logging.debug(f"type: {type(version_rules[0].rules)}")
                    does_query_exist = any(
                        ruleElm.rules.query != None for ruleElm in version_rules
                    )
                    if does_query_exist:
                        deployed_version = next(
                            (
                                v
                                for v in prompt_chain_version_and_rules.versions
                                if v.id == version_id
                            ),
                            None,
                        )
                        if deployed_version:
                            return self.__format_prompt_chain(deployed_version)
            else:
                return self.__format_prompt_chain(
                    prompt_chain_version_and_rules.versions[0]
                )

        if prompt_chain_version_and_rules.fallbackVersion:
            logging.info(
                f"************* FALLBACK TRIGGERED : VERSION : {prompt_chain_version_and_rules.fallbackVersion} *************"
            )
            return self.__format_prompt_chain(
                prompt_chain_version_and_rules.fallbackVersion
            )
        return None

    def __get_folders_for_rule(
        self, folders: List[Folder], rule: QueryRule
    ) -> List[Folder]:
        incoming_query = IncomingQuery(
            query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch
        )
        objects = []
        for folder in folders:
            query = RuleGroupType(rules=[], combinator="AND")
            if not folder.tags:
                continue
            parsed_incoming_query = parse_incoming_query(incoming_query.query)
            tags = folder.tags
            for key, value in tags.items():
                if key in [q.field for q in parsed_incoming_query]:
                    # if not isinstance(value,None):
                    query.rules.append(RuleType(field=key, operator="=", value=value))
            if not query.rules:
                continue
            objects.append(QueryObject(query=query, id=folder.id))

        folder_objects = find_all_matches(objects, incoming_query)
        ids = [fo.id for fo in folder_objects]
        return [f for f in folders if f.id in ids]

    def get_prompt(self, id: str, rule: QueryRule) -> Optional[Prompt]:
        """
        Retrieves a prompt based on the provided id and rule.

        Args:
            id (str): The id of the prompt.
            rule (QueryRule): The rule to match the prompt against.

        Returns:
            Optional[Prompt]: The prompt object if found, otherwise None.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. You can enable it by initializing Maxim with Config(...prompt_management=True)."
            )
        logging.debug(f"getPrompt: id: {id} for rule: {rule}")
        key = self.__get_cache_key("PROMPT", id)
        version_and_rules_with_prompt_id = self.__get_prompt_from_cache(key)
        if version_and_rules_with_prompt_id is None:
            version_and_rules_with_prompt_id = self.maxim_api.get_prompt(id)
            if len(version_and_rules_with_prompt_id.versions) == 0:
                return None
            self.__cache.set(
                id,
                json.dumps(
                    version_and_rules_with_prompt_id,
                    cls=VersionAndRulesWithPromptIdEncoder,
                ),
            )
        if not version_and_rules_with_prompt_id:
            return None
        prompt = self.__get_prompt_version_for_rule(
            version_and_rules_with_prompt_id, rule
        )
        if not prompt:
            return None
        return prompt

    def get_prompts(self, rule: QueryRule) -> List[Prompt]:
        """
        Retrieves all prompts that match the given rule.

        Args:
            rule (QueryRule): The rule to match the prompts against.

        Returns:
            List[Prompt]: A list of prompts that match the given rule.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        version_and_rules = self.__get_all_prompts_from_cache()
        prompts = []
        if version_and_rules is None or len(version_and_rules) == 0:
            self.__sync_entities()
            version_and_rules = self.__get_all_prompts_from_cache()
        if not version_and_rules:
            return []
        for v in version_and_rules:
            if rule.scopes:
                if "folder" not in rule.scopes.keys():
                    return []
                else:
                    if rule.scopes["folder"] != v.folderId:
                        continue
            prompt = self.__get_prompt_version_for_rule(v, rule)
            if prompt is not None:
                prompts.append(prompt)
        if len(prompts) == 0:
            return []
        return prompts

    def get_prompt_chain(self, id: str, rule: QueryRule) -> Optional[PromptChain]:
        """
        Retrieves a prompt chain based on the provided id and rule.

        Args:
            id (str): The id of the prompt chain.
            rule (QueryRule): The rule to match the prompt chain against.

        Returns:
            Optional[PromptChain]: The prompt chain object if found, otherwise None.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config by setting Config(...,prompt_management=True)."
            )
        key = self.__get_cache_key("PROMPT_CHAIN", id)
        version_and_rules = self.__get_prompt_chain_from_cache(key)
        if version_and_rules is None:
            version_and_rules = self.maxim_api.getPromptChain(id)
            if len(version_and_rules.versions) == 0:
                return None
            self.__cache.set(
                id,
                json.dumps(
                    version_and_rules, cls=VersionAndRulesWithPromptChainIdEncoder
                ),
            )
        if not version_and_rules:
            return None
        prompt_chains = self.__get_prompt_chain_version_for_rule(
            version_and_rules, rule
        )
        if not prompt_chains:
            return None
        return prompt_chains

    def get_folder_by_id(self, id: str) -> Optional[Folder]:
        """
        Retrieves a folder based on the provided id.

        Args:
            id (str): The id of the folder.

        Returns:
            Optional[Folder]: The folder object if found, otherwise None.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        key = self.__get_cache_key("FOLDER", id)
        folder = self.__get_folder_from_cache(key)
        if folder is None:
            try:
                folder = self.maxim_api.get_folder(id)
                if not folder:
                    return None
                self.__cache.set(key, json.dumps(folder, cls=FolderEncoder))
            except Exception:
                return None
        return folder

    def get_folders(self, rule: QueryRule) -> List[Folder]:
        """
        Retrieves all folders that match the given rule.

        Args:
            rule (QueryRule): The rule to match the folders against.

        Returns:
            List[Folder]: A list of folders that match the given rule.
        """
        if self.prompt_management is False:
            raise Exception(
                "prompt_management is disabled. Please enable it in config."
            )
        folders = self.__get_all_folders_from_cache()
        if folders is None or len(folders) == 0:
            self.__sync_entities()
            folders = self.__get_all_folders_from_cache()
        if not folders:
            return []
        return self.__get_folders_for_rule(folders, rule)

    def logger(self, config: LoggerConfig) -> Logger:
        """
        Creates a logger based on the provided configuration.

        Args:
            config (LoggerConfig): The configuration for the logger.

        Returns:
            Logger: The logger object.
        """
        # Checking if this log repository exist on server
        exists = self.maxim_api.does_log_repository_exist(config.id)
        if not exists:
            if config.id:
                logging.error("Log repository not found")
                if self.raise_exceptions:
                    raise Exception("Log repository not found")
        if config.id in self.__loggers:
            return self.__loggers[config.id]
        logger = Logger(
            config=config,
            api_key=self.api_key,
            base_url=self.base_url,
            is_debug=self.__is_debug,
            raise_exceptions=self.raise_exceptions,
        )
        self.__loggers[config.id] = logger
        return logger

    def create_test_run(self, name: str, in_workspace_id: str) -> TestRunBuilder:
        """
        Creates a test run builder based on the provided name and workspace id.

        Args:
            name (str): The name of the test run.
            in_workspace_id (str): The workspace id to create the test run in.

        Returns:
            TestRunBuilder: The test run builder object.
        """
        return TestRunBuilder(
            name=name,
            workspace_id=in_workspace_id,
            api_key=self.api_key,
            base_url=self.base_url,
            evaluators=[],
        )

    def cleanup(self):
        """
        Cleans up the Maxim sync thread.
        """
        # We sleep to allow all pending writes in the queue to come in and then we can flush
        time.sleep(2)
        logging.debug("Cleaning up Maxim sync thread")
        self.is_running = False
        for logger in self.__loggers.values():
            logger.cleanup()
        logging.debug("Cleanup done")



import json
from typing import Any, Dict, List, Optional

from .core import (
    validate_content,
    validate_optional_type,
    validate_type,
    validate_type_to_be_one_of,
)


def parse_function_call(function_call_data):
    validate_type(function_call_data.get("name"), str, "name")
    validate_type(function_call_data.get("arguments"), str, "arguments")
    return function_call_data


def parse_tool_calls(tool_calls_data):
    validate_type(tool_calls_data.get("id"), str, "id")
    validate_type(tool_calls_data.get("type"), str, "type")
    parse_function_call(tool_calls_data.get("function"))
    return tool_calls_data


def parse_chat_completion_choice(messages_data):
    validate_type(messages_data.get("role"), str, "role")
    validate_optional_type(messages_data.get("content"), str, "content")
    if messages_data.get("function_call") is not None:
        parse_function_call(messages_data.get("function_call"))
    elif messages_data.get("tool_calls") is not None:
        parse_tool_calls(messages_data.get("tool_calls"))
    return messages_data


def parse_choice(choice_data):
    validate_type(choice_data.get("index"), int, "index")
    validate_type(choice_data.get("finish_reason"), str, "finish_reason")

    # Checking if text completion or chat completion
    if choice_data.get("text") is not None:
        validate_type(choice_data.get("text"), str, "text")
    elif choice_data.get("messages") is not None:
        parse_chat_completion_choice(choice_data.get("messages"))

    return choice_data


def parse_usage(usage_data):
    if usage_data is None:
        return None
    validate_type(usage_data.get("prompt_tokens"), int, "prompt_tokens")
    validate_type(usage_data.get("completion_tokens"), int, "completion_tokens")
    validate_type(usage_data.get("total_tokens"), int, "total_tokens")
    return usage_data


def parse_generation_error(error_data):
    if error_data is None:
        return None
    validate_type(error_data.get("message"), str, "message")
    validate_optional_type(error_data.get("code"), str, "code")
    validate_optional_type(error_data.get("type"), str, "type")
    return error_data


def parse_result(data: Any) -> Dict[str, Any]:
    if not isinstance(data, dict):
        raise ValueError("Text completion is not supported.")
    validate_type(data.get("id"), str, "id")
    validate_optional_type(data.get("object"), str, "object")
    validate_type(data.get("created"), int, "created")
    validate_optional_type(data.get("model"), str, "model")

    choices_data = data.get("choices")
    validate_type_to_be_one_of(choices_data, [list, List], "choices")
    if not choices_data:
        raise ValueError("choices must not be empty")
    choices = [parse_choice(choice) for choice in choices_data]
    usage = parse_usage(data.get("usage", None))
    error = parse_generation_error(data.get("error", None))
    result = {
        "id": data["id"],
        "object": data["object"] if "object" in data else None,
        "created": data["created"],
        "choices": choices,
        "usage": usage,
        "error": error if error else None,
    }
    # removing all None from result
    result = {k: v for k, v in result.items() if v is not None}
    return result


def parse_message(message: Any) -> Any:
    validate_type(message.get("role"), str, "role")
    validate_content(
        message.get("role"), ["user", "assistant", "system", "bot", "chatbot", "model"]
    )
    validate_type_to_be_one_of(message.get("content"), [str, object], "type")
    if isinstance(message.get("content"), object):
        # Making sure if content has type and corresponding data
        content = message.get("content")
        validate_type(content.get("type"), str, "type")
        validate_content(content.get("type"), ["image_url", "text"])
        # Making sure type is image or text
        type = content.get("type")
        if type == "image_url":
            validate_type(content.get("image_url"), str, "image_url")
        elif type == "text":
            validate_type(content.get("text"), str, "text")
        else:
            raise ValueError(
                f"Invalid content type. We expect 'text' or 'image' type. Got: {type}"
            )
    return message


def parse_messages(messages: List[Any]) -> List[Any]:
    if len(messages) == 0:
        return []
    return [parse_message(message) for message in messages]


def parse_model_parameters(parameters: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    # convert parameters dict into JSON string
    if parameters is None:
        return {}
    # we will go through each key and make sure it is a string
    # if not we will do json.dumps on it
    for key, value in parameters.items():
        if not isinstance(value, str):
            parameters[key] = json.dumps(value)
    return parameters



from .feedback import Feedback
from .generation import (
    Generation,
    GenerationConfig,
    GenerationError,
    GenerationRequestMessage,
)
from .retrieval import Retrieval, RetrievalConfig
from .session import Session, SessionConfig
from .span import Span, SpanConfig
from .toolCall import ToolCall, ToolCallConfig, ToolCallError
from .trace import Trace, TraceConfig

__all__ = [
    "Feedback",
    "Generation",
    "GenerationConfig",
    "GenerationError",
    "Retrieval",
    "RetrievalConfig",
    "GenerationRequestMessage",
    "Session",
    "SessionConfig",
    "Span",
    "SpanConfig",
    "Trace",
    "TraceConfig",
    "ToolCall",
    "ToolCallConfig",
    "ToolCallError",
]



import collections.abc
import inspect
import re
import types
from datetime import datetime
from decimal import Decimal
from typing import Any


def make_object_serializable(obj: Any) -> Any:
    """
    Convert any Python object into a JSON-serializable format while preserving
    as much information as possible about the original object.

    Args:
        obj: Any Python object

    Returns:
        A JSON-serializable representation of the object
    """
    # Handle None
    if obj is None:
        return None

    # Handle basic types that are already serializable
    if isinstance(obj, (bool, int, float, str)):
        return obj

    # Handle Decimal
    if isinstance(obj, Decimal):
        return str(obj)

    # Handle complex numbers
    if isinstance(obj, complex):
        return {"type": "complex", "real": obj.real, "imag": obj.imag}

    # Handle bytes and bytearray
    if isinstance(obj, (bytes, bytearray)):
        return {"type": "bytes", "data": obj.hex(), "encoding": "hex"}

    # Handle datetime objects
    if isinstance(obj, datetime):
        return obj.isoformat()

    # Handle regular expressions
    if isinstance(obj, re.Pattern):
        return {"type": "regex", "pattern": obj.pattern, "flags": obj.flags}

    # Handle functions
    if isinstance(obj, (types.FunctionType, types.MethodType)):
        return {
            "type": "function",
            "name": obj.__name__,
            "module": obj.__module__,
            "source": inspect.getsource(obj) if inspect.isroutine(obj) else None,
            "signature": str(inspect.signature(obj))
            if inspect.isroutine(obj)
            else None,
        }

    # Handle exceptions
    if isinstance(obj, Exception):
        return {
            "type": "error",
            "error_type": obj.__class__.__name__,
            "message": str(obj),
            "args": make_object_serializable(obj.args),
            "traceback": str(obj.__traceback__) if obj.__traceback__ else None,
        }

    # Handle sets
    if isinstance(obj, (set, frozenset)):
        return {
            "type": "set",
            "is_frozen": isinstance(obj, frozenset),
            "values": [make_object_serializable(item) for item in obj],
        }

    # Handle dictionaries and mapping types
    if isinstance(obj, collections.abc.Mapping):
        return {str(key): make_object_serializable(value) for key, value in obj.items()}

    # Handle lists, tuples, and other iterables
    if isinstance(obj, (list, tuple)) or (
        isinstance(obj, collections.abc.Iterable)
        and not isinstance(obj, (str, bytes, bytearray))
    ):
        return [make_object_serializable(item) for item in obj]

    # Handle custom objects
    try:
        # Try to convert object's dict representation
        obj_dict = obj.__dict__
        return {
            "type": "custom_object",
            "class": obj.__class__.__name__,
            "module": obj.__class__.__module__,
            "attributes": make_object_serializable(obj_dict),
        }
    except AttributeError:
        # If object doesn't have __dict__, try to get string representation
        return {
            "type": "unknown",
            "class": obj.__class__.__name__,
            "string_repr": str(obj),
        }



from .components.feedback import Feedback
from .components.generation import (
    Generation,
    GenerationConfig,
    GenerationError,
    GenerationRequestMessage,
)
from .components.retrieval import Retrieval, RetrievalConfig
from .components.session import Session, SessionConfig
from .components.span import Span, SpanConfig
from .components.toolCall import ToolCall, ToolCallConfig, ToolCallError
from .components.trace import Trace, TraceConfig
from .logger import Logger, LoggerConfig

__all__ = [
    "Logger",
    "LoggerConfig",
    "TraceConfig",
    "Trace",
    "RetrievalConfig",
    "GenerationRequestMessage",
    "ToolCallError",
    "Retrieval",
    "SpanConfig",
    "Span",
    "GenerationConfig",
    "Generation",
    "SessionConfig",
    "Session",
    "Feedback",
    "ToolCallConfig",
    "ToolCall",
    "GenerationError",
]



import logging
import uuid
from dataclasses import fields
from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID

# Use LangChain features
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.documents import Document
from langchain_core.messages import BaseMessage, ToolMessage
from langchain_core.outputs import LLMResult

from ...expiring_key_value_store import ExpiringKeyValueStore
from ...logger import (
    GenerationConfig,
    Logger,
    RetrievalConfig,
    SpanConfig,
    ToolCallConfig,
    ToolCallError,
)
from ..models import Container, Metadata, SpanContainer, TraceContainer
from .utils import (
    parse_langchain_llm_error,
    parse_langchain_llm_result,
    parse_langchain_messages,
    parse_langchain_model_and_provider,
    parse_langchain_model_parameters,
    parse_langchain_provider,
)

logging = logging.getLogger("MaximSDK")


# 20 minutes
DEFAULT_TIMEOUT = 60 * 20


class MaximLangchainTracer(BaseCallbackHandler):
    """
    A callback handler that logs langchain outputs to Maxim logger

    Args:
        logger: Logger: Maxim Logger instance to log outputs
    """

    def __init__(
        self, logger: Logger, metadata: Optional[Dict[str, Any]] = None
    ) -> None:
        """Initializes the Langchain Tracer
        Args:
            logger: Logger: Maxim Logger instance to log outputs
        """
        super().__init__()
        self.run_inline = True
        self.logger = logger
        self.containers: Dict[str, Container] = {}
        self.metadata_store = ExpiringKeyValueStore()
        self.metadata = None
        if metadata is not None:
            self.__validate_maxim_metadata(metadata)
            self.metadata = metadata

    def __validate_maxim_metadata(self, metadata: Optional[Dict[str, Any]]):
        """
        Validates the metadata
        """
        if metadata is None:
            return
        id_keys = ["session_id", "trace_id", "span_id"]
        present_keys = [key for key in id_keys if key in metadata]
        if len(present_keys) > 1:
            raise ValueError(
                f"Multiple keys found in metadata: {present_keys}. You can pass only one of these."
            )
        valid_keys = [field.name for field in fields(Metadata)]
        invalid_keys = [key for key in metadata if key not in valid_keys]
        if len(invalid_keys) > 0:
            raise ValueError(
                f"Invalid keys found in metadata: {invalid_keys}. Valid keys are {valid_keys}"
            )

    def __get_container_from_metadata(self, run_id: UUID) -> Container:
        """
        Gets the container from the metadata
        """
        maxim_metadata = self.__get_metadata()
        container: Optional[Container] = None
        if maxim_metadata is not None:
            span_id: Optional[str] = maxim_metadata.span_id
            if span_id is not None:
                span_name: Optional[str] = maxim_metadata.span_name
                container = SpanContainer(
                    span_id=span_id,
                    logger=self.logger,
                    span_name=span_name,
                    mark_created=True,
                )
            trace_id = maxim_metadata.trace_id
            if trace_id is not None:
                trace_name = maxim_metadata.trace_name
                container = TraceContainer(
                    trace_id=trace_id,
                    logger=self.logger,
                    trace_name=trace_name,
                    mark_created=True,
                )

        if container is None:
            session_id: Optional[str] = (
                maxim_metadata.session_id if maxim_metadata else None
            )
            container = TraceContainer(
                trace_id=str(uuid.uuid4()),
                logger=self.logger,
                trace_name="Trace",
                parent=session_id,
            )
            container.create()

        return container

    def __get_container(
        self, run_id: UUID, parent_run_id: Optional[UUID] = None
    ) -> Optional[Container]:
        container: Optional[Container] = None
        if parent_run_id is None:
            # This is the first activity in this run
            container = self.__get_container_from_metadata(run_id)
        else:
            container = self.containers.get(str(parent_run_id))
        return container

    def __get_metadata(
        self, metadata: Optional[Dict[str, Any]] = None
    ) -> Optional[Metadata]:
        if metadata is not None and metadata.get("maxim", None) is not None:
            return Metadata(metadata.get("maxim", None))
        return Metadata(self.metadata)

    def on_llm_start(
        self,
        serialized: dict[str, Any],
        prompts: list[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Runs when LLM starts"""
        logging.debug("[MaximSDK: Langchain] on_llm_start called")
        model, model_parameters = parse_langchain_model_parameters(**kwargs)
        provider = parse_langchain_provider(serialized)
        model, provider = parse_langchain_model_and_provider(model, provider)
        messages = parse_langchain_messages(prompts)
        generation_id = str(run_id)
        maxim_metadata = self.__get_metadata(metadata)
        # Prepare generation config
        generation_config = GenerationConfig(
            id=generation_id,
            name=maxim_metadata.generation_name if maxim_metadata else None,
            provider=provider,
            model=model,
            messages=messages,
            model_parameters=model_parameters,
            tags=maxim_metadata.generation_tags if maxim_metadata else None,
        )
        # Adding it back to the container
        container = self.__get_container(run_id, parent_run_id)
        if container is None:
            logging.error("[MaximSDK] Couldn't find a container for generation")
            return
        if not container.is_created():
            container.create()
        container.add_generation(generation_config)

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Runs when a chat model call starts"""
        logging.debug("[Langchain] on_chat_model_start called")
        run_id = kwargs.get("run_id", None)
        model, model_parameters = parse_langchain_model_parameters(**kwargs)
        provider = parse_langchain_provider(serialized)
        model, provider = parse_langchain_model_and_provider(model, provider)
        maxim_messages = parse_langchain_messages(messages)
        generation_id = str(run_id)
        maxim_metadata = self.__get_metadata(metadata)
        generation_name = None
        if kwargs.get("parent_run_id", None) is None:
            generation_name = maxim_metadata.generation_name if maxim_metadata else None
        # Checking if generation id already present for this run_id
        generation_config = GenerationConfig(
            id=generation_id,
            name=generation_name,
            provider=provider,
            model=model,
            messages=maxim_messages,
            model_parameters=model_parameters,
            tags=maxim_metadata.generation_tags if maxim_metadata else None,
        )
        container = self.__get_container(run_id, kwargs.get("parent_run_id", None))
        if container is None:
            logging.error("[MaximSDK] Couldn't find a container for  generation")
            return
        if not container.is_created():
            container.create()
        container.add_generation(generation_config)
        if container.parent() is None:
            self.containers[str(run_id)] = container

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        logging.debug("[MaximSDK: Langchain] on_llm_end called")
        run_id = kwargs.get("run_id", None)
        parent_run_id = kwargs.get("parent_run_id", None)
        result = parse_langchain_llm_result(response)
        container = self.__get_container(run_id, parent_run_id)
        if container is None:
            logging.error("[MaximSDK] Couldn't find a container for generation")
            return
        self.logger.generation_result(str(run_id), result)
        # We close the container - only if it was created by this run_id
        if container.parent() is None:
            container.end()
            self.containers.pop(str(run_id))

    def on_llm_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when LLM errors."""
        logging.debug("[Langchain] on_llm_error called")
        run_id = kwargs.get("run_id", None)
        parent_run_id = kwargs.get("parent_run_id", None)
        container = self.__get_container(run_id, parent_run_id)
        if container is None:
            logging.error("[MaximSDK] Couldn't find a container for generation")
            return
        generation_error = parse_langchain_llm_error(error)
        self.logger.generation_error(str(run_id), generation_error)
        # We close the container - only if it was created by this run_id
        if container.parent() is None:
            container.end()
            self.containers.pop(str(run_id))

    def on_retriever_start(
        self,
        serialized: dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Run when Retriever starts running.
        """
        logging.debug("[Langchain] on_retriever_start called")
        retrieval_id = str(run_id)
        retrieval_config = RetrievalConfig(id=retrieval_id)
        container = self.__get_container(run_id, parent_run_id)
        if container is None:
            logging.error("[MaximSDK] Couldn't find a container for retrieval")
            return
        if not container.is_created():
            container.create()
        retrieval = container.add_retrieval(retrieval_config)
        retrieval.input(query)

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logging.debug("[Langchain] on_retriever_end called")
        container = self.__get_container(run_id, parent_run_id)
        if container is None:
            logging.error("[MaximSDK] Couldn't find a container for retrieval")
            return
        self.logger.retrieval_output(str(run_id), documents)
        if container.parent is None:
            container.end()
            self.containers.pop(str(run_id))

    # Chain callbacks

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            # Creating new span for the chain node
            name = kwargs.get("name", None)
            tags: Dict[str, str] = {}
            if (
                self.metadata is not None
                and self.metadata.get("chain_tags") is not None
            ):
                # its a dict
                for key, value in self.metadata["chain_tags"].items():
                    tags[key] = str(value)
            tags["run_id"] = run_id
            if kwargs.get("parent_run_id", None) is not None:
                tags["parent_run_id"] = str(kwargs.get("parent_run_id", None))
            if kwargs.get("tags", None) is not None:
                for tag in kwargs["tags"]:
                    key, value = tag.split(":", 1)
                    tags[key.strip()] = value.strip()
            if kwargs.get("metadata", None) is not None:
                for key, value in kwargs["metadata"].items():
                    tags[key.strip()] = str(value)

            container = self.__get_container(run_id, kwargs.get("parent_run_id", None))

            if container is None:
                logging.error("[MaximSDK] Couldn't find a container for chain")
                return

            if not container.is_created():
                container.create()

            span_config = SpanConfig(id=str(run_id), name=name, tags=tags)
            container.add_span(span_config)
            self.containers[str(run_id)] = SpanContainer(
                span_id=str(run_id), logger=self.logger, mark_created=True
            )
        except Exception as e:
            logging.error("[MaximSDK] Error while processing chain_start %s", e)

    def on_chain_end(self, outputs: Union[str, Dict[str, Any]], **kwargs: Any) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            # We hide the hidden chains
            tags = {
                tag.split(":")[0]: tag.split(":")[1]
                for tag in kwargs.get("tags", [])
                if ":" in tag
            }
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                logging.error("[MaximSDK] Couldn't find a container for chain")
                return

            container.add_tags(tags)
            container.end()
            self.containers.pop(str(run_id))

        except Exception as e:
            logging.error("[MaximSDK] Failed to parse chain-end: %s", e)

    def on_chain_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        logging.info("[MaximSDK] on_chain_error called")

    def on_custom_event(
        self,
        name: str,
        data: Any,
        *,
        run_id: UUID,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        container = self.__get_container(run_id, kwargs.get("parent_run_id", None))
        if container is None:
            logging.error("[MaximSDK] Couldn't find a container for event")
            return
        if tags is not None:
            final_tags = {
                tag.split(":")[0]: tag.split(":")[1] for tag in tags if ":" in tag
            }
            container.add_event(str(run_id), name, final_tags)

    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logging.warning("[MaximSDK] On agent action is not supported")

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logging.warning("[MaximSDK] On agent action is not supported")

    # Tool callbacks

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            name = serialized.get("name", None)
            description = serialized.get("description", None)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                logging.error("[MaximSDK] Couldn't find a container for this tool call")
                return
            if not container.is_created():
                container.create()
            container.add_tool_call(
                ToolCallConfig(
                    id=str(run_id), name=name, description=description, args=input_str
                )
            )
            if container.parent() is None:
                self.containers[str(run_id)] = container
        except Exception as e:
            logging.error("[MaximSDK] Failed to parse tool-start: %s", e)

    def on_tool_end(self, output: Any, **kwargs: Any) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            parent_run_id = kwargs.get("parent_run_id", None)
            container = self.__get_container(run_id, parent_run_id)
            if container is None:
                logging.error("[MaximSDK] Couldn't find a container for tool call")
                return
            if isinstance(output, ToolMessage):
                if output.status == "success":
                    self.logger.tool_call_result(str(run_id), output.content)
                elif output.status == "error":
                    if isinstance(output.content, str):
                        self.logger.tool_call_error(
                            str(run_id), ToolCallError(message=output.content)
                        )
                    else:
                        self.logger.tool_call_error(
                            str(run_id), ToolCallError(message=str(output.content))
                        )
            if container.parent() is None:
                container.end()
                self.containers.pop(str(run_id))
        except Exception as e:
            logging.error("[MaximSDK] Failed to parse tool-end: %s", e)

    def on_tool_error(
        self, error: Union[Exception, BaseException, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        try:
            run_id = kwargs.get("run_id", None)
            container = self.__get_container(run_id, kwargs.get("parent_run_id", None))
            if container is None:
                logging.error("[MaximSDK] Couldn't find a container for tool_call")
                return
            if error is not None and hasattr(error, "__str__"):
                self.logger.tool_call_error(
                    str(run_id), ToolCallError(message=str(error))
                )
            if container.parent() is None:
                container.end()
                self.containers.pop(str(run_id))
        except Exception as e:
            logging.error("[MaximSDK] Failed to parse tool-end: %s", e)

    def on_text(self, text: str, **kwargs: Any) -> Any:
        logging.warning("[MaximSDK] Text models are not supported by Maxim callback")



import json
import logging
import re
import time
from typing import Any, Dict, List, Tuple, Union
from uuid import uuid4

from langchain_core.messages import AIMessage, BaseMessage, ToolCall
from langchain_core.outputs import LLMResult
from langchain_core.outputs.chat_generation import ChatGeneration
from langchain_core.outputs.generation import Generation

from ..components.types import GenerationError

logger = logging.getLogger("MaximSDK")


def parse_langchain_model_and_provider(model: str, provider: str) -> Tuple[str, str]:
    return model, provider


def parse_langchain_provider(serialized: Dict[str, Any]):
    """Parses langchain provider from serialized data
    Args:
        serialized: Dict[str, Any]: Serialized data to parse provider from
    Returns:
        str: Parsed provider
    """
    provider = serialized.get("name", "").lower()
    if provider.startswith("chat"):
        return provider.replace("chat", "")
    elif "azure" in provider:
        return "azure"
    elif "anthropic" in provider:
        return "anthropic"
    elif "huggingface" in provider:
        return "huggingface"
    elif "bedrock" in provider:
        return "bedrock"
    elif "aws" in provider:
        return "bedrock"
    return provider


def parse_langchain_llm_error(
    error: Union[Exception, BaseException, KeyboardInterrupt],
) -> GenerationError:
    """Parses langchain LLM error into a format that is accepted by Maxim logger
    Args:
        error: Union[Exception, KeyboardInterrupt]: Error to be parsed
    Returns:
        GenerationError: Parsed LLM error
    """
    if isinstance(error, KeyboardInterrupt):
        return GenerationError(message="Generation was interrupted by the user")
    if isinstance(error, Exception):
        return GenerationError(message=str(error))
    else:
        message = error.__dict__.get("message", "")
        type = error.__dict__.get("type", None)
        code = error.__dict__.get("code", None)
        return GenerationError(message=message, type=type, code=code)


def parse_langchain_model_parameters(**kwargs: Any) -> Tuple[str, Dict[str, Any]]:
    """Parses langchain kwargs into model and model parameters. You can use this function with any langchain _start callback function
    Args:
        kwargs: Dict[str, Any]: Kwargs to be parsed
    Returns:
        Tuple[str, Dict[str, Any]]: Model and model parameters
    Raises:
        Exception: If model_name is not found in kwargs
    """
    model_parameters = kwargs.get("invocation_params", {})
    # Checking if model_name present
    model = "unknown"
    if "model_name" in model_parameters:
        model = model_parameters["model_name"]
        del model_parameters["model_name"]
    # If not then checking if invocation_params exist
    elif "model" in model_parameters:
        model = model_parameters["model"]
        del model_parameters["model"]
    elif "model_id" in model_parameters:
        model = model_parameters["model_id"]
        del model_parameters["model_id"]
    return model, model_parameters


def parse_base_message_to_maxim_generation(message: BaseMessage):
    """Parses langchain BaseMessage into a format that is accepted by Maxim logger
    Args:
        message: BaseMessage
    Returns:
        Dict[str, Any]: Parsed message
    """
    choice = parse_langchain_message(message)
    usage = (
        message.__dict__["usage_metadata"] if message.__dict__["usage_metadata"] else {}
    )
    return {
        "id": str(uuid4()),
        "created": int(time.time()),
        "choices": [choice],
        "usage": {
            "prompt_tokens": usage.get("input_tokens", 0),
            "completion_tokens": usage.get("output_tokens", 0),
            "total_tokens": usage.get("total_tokens", 0),
        },
    }


def parse_langchain_message(message: BaseMessage):
    """Parses langchain BaseMessage into a choice of openai message
    Args:
        message: BaseMessage
    Returns:
        Dict[str, Any]: Parsed message
    """
    message_type = message.__dict__["type"]
    if message_type == "ai":
        ai_message = AIMessage(
            content=message.content, additional_kwargs=message.additional_kwargs
        )
        tool_calls = (
            ai_message.tool_calls or message.lc_attributes.get("tool_calls") or []
        )
        return {
            "index": 0,
            "message": {
                "role": "assistant",
                "content": (
                    message.content
                    if isinstance(message.content, str)
                    else json.dumps(message.content)
                ),
                "tool_calls": parse_langchain_tool_call(tool_calls),
            },
            "finish_reason": (
                message.response_metadata.get("finish_reason")
                or message.response_metadata.get("stop_reason")
                or "stop"
            ),
            "logprobs": (message.response_metadata.get("logprobs") or None),
        }
    else:
        return {
            "index": 0,
            "message": {
                "role": "assistant",
                "content": message.content,
            },
            "finish_reason": message.response_metadata["finish_reason"]
            if message.response_metadata["finish_reason"]
            else None,
            "logprobs": message.response_metadata["logprobs"]
            if message.response_metadata["logprobs"]
            else None,
        }


def parse_langchain_tool_call(tool_calls: List[ToolCall]):
    final_tool_calls = []
    for tool_call in tool_calls:
        final_tool_calls.append(
            {
                "type": "function",
                "id": tool_call.get("id"),
                "function": {
                    "name": tool_call.get("name"),
                    "arguments": json.dumps(tool_call.get("args")),
                },
            }
        )
    return final_tool_calls


def parse_langchain_chat_generation_chunk(generation: ChatGeneration):
    choices = []
    content = generation.text
    finish_reason = (
        generation.message.response_metadata.get("stop_reason")
        if generation.message.response_metadata
        else "stop"
    )
    if finish_reason is None:
        finish_reason = (
            generation.generation_info.get("finish_reason")
            if generation.generation_info
            else "stop"
        )
    choices.append(
        {
            "index": 0,
            "message": {"role": "assistant", "content": content, "tool_calls": []},
            "finish_reason": finish_reason,
            "logprobs": generation.generation_info.get("logprobs")
            if generation.generation_info
            else None,
        }
    )
    return choices


def parse_langchain_chat_generation(generation: ChatGeneration):
    choices = []
    message = generation.message
    if message.type == "ai":
        ai_message = AIMessage(
            content=message.content, additional_kwargs=message.additional_kwargs
        )
        finish_reason = (
            generation.message.response_metadata.get("stop_reason")
            if generation.message.response_metadata
            else "stop"
        )
        if finish_reason is None:
            finish_reason = (
                generation.generation_info.get("finish_reason")
                if generation.generation_info
                else "stop"
            )
        tool_calls = (
            ai_message.tool_calls or message.lc_attributes.get("tool_calls") or []
        )
        choices.append(
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": ai_message.content,
                    "tool_calls": parse_langchain_tool_call(tool_calls),
                },
                "finish_reason": finish_reason,
                "logprobs": generation.generation_info.get("logprobs")
                if generation.generation_info
                else None,
            }
        )
    return choices


def parse_langchain_text_generation(generation: Generation):
    choices = []
    messages = parse_langchain_messages([generation.text], "system")
    if len(messages) > 0:
        for i, message in enumerate(messages):
            choices.append(
                {
                    "index": i,
                    "text": message["content"],
                    "logprobs": generation.generation_info.get("logprobs")
                    if generation.generation_info
                    else None,
                    "finish_reason": generation.generation_info.get("finish_reason")
                    if generation.generation_info
                    else None,
                }
            )
    return choices


def parse_langchain_generation(generation: Generation):
    """Parses langchain generation into a format that is accepted by Maxim logger
    Args:
        generation: Generation: Generation to be parsed
    Returns:
        Dict[str, Any]: Parsed generation
    """
    if generation.type == "ChatGeneration":
        return parse_langchain_chat_generation(generation)
    elif generation.type == "ChatGenerationChunk":
        return parse_langchain_chat_generation_chunk(generation)
    elif generation.type == "Generation":
        return parse_langchain_text_generation(generation)
    else:
        raise Exception(f"Invalid generation type: {generation.type}")


def parse_token_usage_for_result(result: LLMResult):
    """
    Parses token usage for a given LLM result
    """
    usage = result.llm_output.get("token_usage") if result.llm_output else None
    if usage is not None:
        return usage
    llm_usage = result.llm_output.get("usage") if result.llm_output else None
    if llm_usage:
        if llm_usage.get("input_tokens") is not None:
            usage = {
                "prompt_tokens": llm_usage.get("input_tokens", 0),
                "completion_tokens": llm_usage.get("output_tokens", 0),
                "total_tokens": llm_usage.get("input_tokens", 0)
                + llm_usage.get("output_tokens", 0),
            }
        elif llm_usage.get("prompt_tokens") is not None:
            usage = {
                "prompt_tokens": llm_usage.get("prompt_tokens", 0),
                "completion_tokens": llm_usage.get("completion_tokens", 0),
                "total_tokens": llm_usage.get("prompt_tokens", 0)
                + llm_usage.get("completion_tokens", 0),
            }
    if usage is not None:
        return usage
    # Here we might have to go down to each generation and sum up all usages
    prompt_tokens = 0
    output_tokens = 0
    total_tokens = 0
    generations = result.generations
    if generations is not None:
        for _, generation in enumerate(generations):
            for _, gen in enumerate(generation):
                usage_data = gen.__dict__.get("message").__dict__.get("usage_metadata")
                if usage_data is not None:
                    if usage_data.get("input_tokens") is not None:
                        prompt_tokens += usage_data.get("input_tokens", 0)
                        output_tokens += usage_data.get("output_tokens", 0)
                        total_tokens += usage_data.get(
                            "input_tokens", 0
                        ) + usage_data.get("output_tokens", 0)
                        continue
                    elif usage_data.get("prompt_tokens") is not None:
                        prompt_tokens += usage_data.get("prompt_tokens", 0)
                        output_tokens += usage_data.get("completion_tokens", 0)
                        total_tokens += usage_data.get(
                            "prompt_tokens", 0
                        ) + usage_data.get("completion_tokens", 0)
                        continue
                resp_metadata = gen.__dict__.get("message").__dict__.get(
                    "response_metadata"
                )
                if resp_metadata is not None:
                    usage_data = resp_metadata.get("usage") or None
                    if usage_data is not None:
                        if usage_data.get("input_tokens") is not None:
                            prompt_tokens += usage_data.get("input_tokens", 0)
                            output_tokens += usage_data.get("output_tokens", 0)
                            total_tokens += usage_data.get(
                                "input_tokens", 0
                            ) + usage_data.get("output_tokens", 0)
                            continue
                        elif usage_data.get("prompt_tokens") is not None:
                            prompt_tokens += usage_data.get("prompt_tokens", 0)
                            output_tokens += usage_data.get("completion_tokens", 0)
                            total_tokens += usage_data.get(
                                "prompt_tokens", 0
                            ) + usage_data.get("completion_tokens", 0)
                            continue
                    # The last case we check for is bedrock
                    usage_data = (
                        resp_metadata.get("amazon-bedrock-invocationMetrics")
                        if resp_metadata
                        else None
                    )
                    if usage_data is not None:
                        prompt_tokens += usage_data.get("inputTokenCount", 0)
                        output_tokens += usage_data.get("outputTokenCount", 0)
                        total_tokens += usage_data.get(
                            "inputTokenCount", 0
                        ) + usage_data.get("outputTokenCount", 0)
    return {
        "prompt_tokens": prompt_tokens,
        "completion_tokens": output_tokens,
        "total_tokens": total_tokens,
    }


def parse_langchain_llm_result(result: LLMResult) -> Dict[str, Any]:
    """Parses langchain LLM result into a format that is accepted by Maxim logger
    Args:
        result: LLMResult: LLM result to be parsed
    Returns:
        Dict[str, Any]: Parsed LLM result
    Raises:
        Exception: If error parsing LLM result
    """
    try:
        generations = result.generations
        choices = []
        if generations is not None:
            for _, generation in enumerate(generations):
                for _, gen in enumerate(generation):
                    parsed_generations = parse_langchain_generation(gen)
                    if isinstance(parsed_generations, list):
                        choices.extend(parsed_generations)
                    else:
                        choices.append(parsed_generations)
        usage = parse_token_usage_for_result(result)
        # Adding index to each choice
        for i, choice in enumerate(choices):
            choices[i] = {**choice, "index": i}
        return {
            "id": str(uuid4()),
            "created": int(time.time()),
            "choices": choices,
            "usage": usage,
        }
    except Exception as e:
        logger.error(f"Error parsing LLM result: {str(e)}")
        raise Exception(f"Error parsing LLM result: {str(e)}")


def parse_langchain_messages(
    input: Union[List[str], List[List[Any]]], default_role="user"
):
    """Parses langchain messages into messages that are accepted by Maxim logger
    Args:
        input: List[str] or List[List[Any]]: List of messages to be parsed
        default_role: str: Default role to assign to messages without a role
    Returns:
        List[Dict[str, str]]: List of messages with role and content
    Raises:
        Exception: If input is not List[str] or List[List[Any]]
        Exception: If message type is not str or list
        Exception: If message type is not recognized
    """
    try:
        delimiter_to_role = {
            "System": "system",
            "Human": "user",
            "User": "user",
            "Assistant": "assistant",
            "Model": "model",
        }
        messages = []
        # checking if input is List[str] or List[List]
        if isinstance(input[0], list):
            for message_list in input or []:
                for message in message_list:
                    if isinstance(message, str):
                        continue
                    message_type = type(message).__name__
                    if message_type.endswith("SystemMessage"):
                        messages.append(
                            {"role": "system", "content": message.content or ""}
                        )
                    elif message_type.endswith("HumanMessage"):
                        messages.append(
                            {"role": "user", "content": message.content or ""}
                        )
                    elif message_type.endswith("AIMessage"):
                        messages.append(
                            {"role": "assistant", "content": message.content or ""}
                        )
                    elif message_type.endswith("ToolMessage"):
                        messages.append(
                            {"role": "user", "content": message.content or ""}
                        )
                    else:
                        logger.error(f"Invalid message type: {type(message)}")
                        raise Exception(f"Invalid message type: {type(message)}")
        else:
            for message in input or []:
                if not isinstance(message, str):
                    logger.error(f"Invalid message type: {type(message)}")
                    raise Exception(f"Invalid message type: {type(message)}")
                # get type of message
                # Define the delimiter pattern
                pattern = r"(System:|Human:|User:|Assistant:|Model:)"
                # Split the text using the pattern
                splits = re.split(pattern, message)
                # Remove any leading/trailing whitespace and empty strings
                splits = [s.strip() for s in splits if s.strip()]
                # Pair up the delimiters with their content
                for i in range(0, len(splits), 2):
                    if i + 1 < len(splits):
                        # Remove ":" from delimiter and trim both delimiter and content
                        delimiter = splits[i].rstrip(":").strip()
                        content = splits[i + 1].strip()
                        messages.append(
                            {
                                "role": delimiter_to_role.get(delimiter, "user"),
                                "content": content,
                            }
                        )
                    else:
                        if splits[i].find(":") == -1:
                            messages.append(
                                {
                                    "role": delimiter_to_role.get(default_role, "user"),
                                    "content": splits[i],
                                }
                            )
                        else:
                            # Handle case where there's a delimiter without content
                            delimiter = splits[i].rstrip(":").strip()
                            messages.append(
                                {
                                    "role": delimiter_to_role.get(delimiter, "user"),
                                    "content": "",
                                }
                            )
        return messages
    except Exception as e:
        logger.error(f"Error parsing messages: {e}")
        raise Exception(f"Error parsing messages: {e}")



import contextlib
import json
from typing import Any, Dict, List, Optional, Union

import requests
from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager
from urllib3.util import Retry, Timeout

from ..models import (
    DatasetEntry,
    DatasetRow,
    Evaluator,
    ExecutePromptForDataResponse,
    ExecuteWorkflowForDataResponse,
    Folder,
    HumanEvaluationConfig,
    RunType,
    TestRun,
    TestRunEntry,
    TestRunResult,
    TestRunStatus,
    TestRunWithDatasetEntry,
    VersionAndRulesWithPromptChainId,
    VersionAndRulesWithPromptId,
)


class ConnectionPool:
    def __init__(self):
        self.session = requests.Session()
        retries = Retry(
            connect=5,
            read=3,
            redirect=1,
            status=3,
            backoff_factor=0.4,
            status_forcelist=frozenset({413, 429, 500, 502, 503, 504}),
        )
        self.http = PoolManager(
            num_pools=2,
            maxsize=3,
            retries=retries,
            timeout=Timeout(connect=10, read=10),
        )
        self.session.mount("https://", HTTPAdapter(max_retries=retries))
        self.session.mount("http://", HTTPAdapter(max_retries=retries))

    @contextlib.contextmanager
    def get_session(self):
        try:
            yield self.session
        finally:
            self.session.close()

    @contextlib.contextmanager
    def get_connection(self):
        try:
            yield self.http
        finally:
            self.http.clear()


class MaximAPI:
    connection_pool: ConnectionPool

    def __init__(self, base_url: str, api_key: str):
        self.connection_pool = ConnectionPool()
        self.base_url = base_url
        self.api_key = api_key

    def __make_network_call(
        self,
        method: str,
        endpoint: str,
        body: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> bytes:
        if headers is None:
            headers = {}
        headers["x-maxim-api-key"] = self.api_key
        url = f"{self.base_url}{endpoint}"
        with self.connection_pool.get_session() as session:
            response = session.request(method, url, data=body, headers=headers)
            response.raise_for_status()
            return response.content

    def get_prompt(self, id: str) -> VersionAndRulesWithPromptId:
        res = self.__make_network_call(
            method="GET", endpoint=f"/api/sdk/v4/prompts?promptId={id}"
        )
        data = json.loads(res.decode())["data"]
        return VersionAndRulesWithPromptId.from_dict(data)

    def get_prompts(self) -> List[VersionAndRulesWithPromptId]:
        res = self.__make_network_call(method="GET", endpoint="/api/sdk/v4/prompts")
        return [
            VersionAndRulesWithPromptId.from_dict(data)
            for data in json.loads(res)["data"]
        ]

    def getPromptChain(self, id: str) -> VersionAndRulesWithPromptChainId:
        res = self.__make_network_call(
            method="GET", endpoint=f"/api/sdk/v4/prompt-chains?promptChainId={id}"
        )
        json_response = json.loads(res.decode())
        return VersionAndRulesWithPromptChainId(**json_response["data"])

    def get_prompt_chains(self) -> List[VersionAndRulesWithPromptChainId]:
        res = self.__make_network_call(
            method="GET", endpoint="/api/sdk/v4/prompt-chains"
        )
        json_response = json.loads(res.decode())
        return [
            VersionAndRulesWithPromptChainId(**elem) for elem in json_response["data"]
        ]

    def get_folder(self, id: str) -> Folder:
        res = self.__make_network_call(
            method="GET", endpoint=f"/api/sdk/v3/folders?folderId={id}"
        )
        json_response = json.loads(res.decode())
        if "tags" not in json_response:
            json_response["tags"] = {}
        return Folder(**json_response["data"])

    def get_folders(self) -> List[Folder]:
        res = self.__make_network_call(method="GET", endpoint="/api/sdk/v3/folders")
        json_response = json.loads(res.decode())
        for elem in json_response["data"]:
            if "tags" not in elem:
                elem["tags"] = {}
        return [Folder(**elem) for elem in json_response["data"]]

    def add_dataset_entries(
        self, dataset_id: str, dataset_entries: List[DatasetEntry]
    ) -> dict:
        res = self.__make_network_call(
            method="POST",
            endpoint="/api/sdk/v3/datasets/entries",
            body=json.dumps(
                {
                    "datasetId": dataset_id,
                    "entries": [entry.to_json() for entry in dataset_entries],
                }
            ),
        )
        return json.loads(res.decode())

    def get_dataset_total_rows(self, dataset_id: str) -> int:
        try:
            res = self.__make_network_call(
                method="GET",
                endpoint=f"/api/sdk/v1/datasets/total-rows?datasetId={dataset_id}",
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return json_response["data"]
        except Exception as e:
            raise Exception(e)

    def get_dataset_row(self, dataset_id: str, row_index: int) -> Optional[DatasetRow]:
        try:
            res = self.__make_network_call(
                method="GET",
                endpoint=f"/api/sdk/v2/datasets/row?datasetId={dataset_id}&row={row_index}",
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return DatasetRow.dict_to_class(json_response["data"])
        except Exception as e:
            if type(e) is requests.exceptions.HTTPError:
                if e.response.status_code == 404:
                    return None
            raise Exception(e)

    def get_dataset_structure(self, dataset_id: str) -> Dict[str, str]:
        try:
            res = self.__make_network_call(
                method="GET",
                endpoint=f"/api/sdk/v1/datasets/structure?datasetId={dataset_id}",
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return json_response["data"]
        except Exception as e:
            raise Exception(e)

    def does_log_repository_exist(self, logger_id: str) -> bool:
        try:
            res = self.__make_network_call(
                method="GET",
                endpoint=f"/api/sdk/v3/log-repositories?loggerId={logger_id}",
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                return False
            return True
        except Exception:
            return False

    def push_logs(self, repository_id: str, logs: str) -> None:
        try:
            res = self.__make_network_call(
                method="POST",
                endpoint=f"/api/sdk/v3/log?id={repository_id}",
                body=logs,
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
        except Exception as e:
            raise Exception(e)

    def fetch_platform_evaluator(self, name: str, in_workspace_id: str) -> Evaluator:
        try:
            res = self.__make_network_call(
                method="GET",
                endpoint=f"/api/sdk/v1/evaluators?name={name}&workspaceId={in_workspace_id}",
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return Evaluator.dict_to_class(json_response["data"])
        except Exception as e:
            raise Exception(e)

    def create_test_run(
        self,
        name: str,
        workspace_id: str,
        workflow_id: Optional[str],
        prompt_version_id: Optional[str],
        run_type: RunType,
        evaluator_config: list[Evaluator],
        human_evaluation_config: Optional[HumanEvaluationConfig] = None,
    ) -> TestRun:
        try:
            res = self.__make_network_call(
                method="POST",
                endpoint="/api/sdk/v1/test-run/create",
                body=json.dumps(
                    {
                        k: v
                        for k, v in {
                            "name": name,
                            "workspaceId": workspace_id,
                            "runType": run_type.value,
                            "workflowId": (
                                workflow_id if workflow_id is not None else None
                            ),
                            "promptVersionId": (
                                prompt_version_id
                                if prompt_version_id is not None
                                else None
                            ),
                            "evaluatorConfig": [
                                evaluator.to_dict() for evaluator in evaluator_config
                            ],
                            "humanEvaluationConfig": (
                                human_evaluation_config.to_dict()
                                if human_evaluation_config
                                else None
                            ),
                        }.items()
                        if v is not None
                    }
                ),
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return TestRun.dict_to_class(json_response["data"])
        except Exception as e:
            raise Exception(e)

    def attach_dataset_to_test_run(self, test_run_id: str, dataset_id: str) -> None:
        try:
            res = self.__make_network_call(
                method="POST",
                endpoint="/api/sdk/v1/test-run/attach-dataset",
                body=json.dumps({"testRunId": test_run_id, "datasetId": dataset_id}),
                headers={"Content-Type": "application/json"},
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
        except Exception as e:
            raise Exception(e)

    def push_test_run_entry(
        self,
        test_run: Union[TestRun, TestRunWithDatasetEntry],
        entry: TestRunEntry,
        run_config: Optional[Dict[str, Any]],
    ) -> None:
        try:
            # making sure run_config has not null values
            if run_config is not None:
                run_config = {k: v for k, v in run_config.items() if v is not None}
            res = self.__make_network_call(
                method="POST",
                endpoint="/api/sdk/v1/test-run/push",
                body=json.dumps(
                    {
                        "testRun": test_run.to_dict(),
                        **({"runConfig": run_config} if run_config is not None else {}),
                        "entry": entry.to_dict(),
                    }
                ),
                headers={"Content-Type": "application/json"},
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
        except Exception as e:
            raise Exception(e)

    def mark_test_run_processed(self, test_run_id: str) -> None:
        try:
            res = self.__make_network_call(
                method="POST",
                endpoint="/api/sdk/v1/test-run/mark-processed",
                body=json.dumps({"testRunId": test_run_id}),
                headers={"Content-Type": "application/json"},
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
        except Exception as e:
            raise Exception(e)

    def get_test_run_status(self, test_run_id: str) -> TestRunStatus:
        try:
            res = self.__make_network_call(
                method="GET",
                endpoint=f"/api/sdk/v1/test-run/status?testRunId={test_run_id}",
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            status: Dict[str, Any] = {}
            status = json_response["data"]["entryStatus"]
            status["testRunStatus"] = json_response["data"]["testRunStatus"]
            return TestRunStatus.dict_to_class(status)
        except Exception as e:
            raise Exception(e)

    def get_test_run_final_result(self, test_run_id: str) -> TestRunResult:
        try:
            res = self.__make_network_call(
                method="GET",
                endpoint=f"/api/sdk/v1/test-run/result?testRunId={test_run_id}",
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return TestRunResult.dict_to_class(json_response["data"])
        except requests.HTTPError as e:
            if e.response is not None and e.response.json() is not None:
                error = e.response.json()
                raise Exception(error["error"])
            raise Exception(e)

    def execute_workflow_for_data(
        self,
        workflow_id: str,
        data_entry: Dict[str, Union[str, List[str], None]],
    ) -> ExecuteWorkflowForDataResponse:
        try:
            res = self.__make_network_call(
                method="POST",
                endpoint="/api/sdk/v1/test-run/execute/workflow",
                body=json.dumps(
                    {
                        "workflowId": workflow_id,
                        "dataEntry": data_entry,
                    }
                ),
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return ExecuteWorkflowForDataResponse.dict_to_class(json_response["data"])
        except requests.HTTPError as e:
            if e.response is not None and e.response.json() is not None:
                error = e.response.json()
                raise Exception(error["error"])
            raise Exception(e)

    def execute_prompt_for_data(
        self,
        prompt_version_id: str,
        input: str,
        data_entry: Dict[str, Union[str, List[str], None]],
    ) -> ExecutePromptForDataResponse:
        try:
            res = self.__make_network_call(
                method="POST",
                endpoint="/api/sdk/v1/test-run/execute/prompt",
                body=json.dumps(
                    {
                        "promptVersionId": prompt_version_id,
                        "input": input,
                        "dataEntry": data_entry,
                    }
                ),
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
            )
            json_response = json.loads(res.decode())
            if "error" in json_response:
                raise Exception(json_response["error"])
            return ExecutePromptForDataResponse.dict_to_class(json_response["data"])
        except requests.HTTPError as e:
            if e.response is not None and e.response.json() is not None:
                error = e.response.json()
                raise Exception(error["error"])
            raise Exception(e)
